/*
 *   bit_utils is a support module which provides bit oriented utilities
 *   Copyright (C) 2013 Elio Blanca <eblanca76@users.sourceforge.net>
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef BIT_UTILS_H
#define BIT_UTILS_H

/*
 * you can take 8*511 bit
 * from the bit reservoir and at most 8*1440 bit from the current
 * frame (320 kbps, 32 kHz), so 8*1951 bit is the largest possible
 * value for MPEG1 and 2)
 * UPDATE: the biggest supported frame is now 8*5760 bit due to a free
 * format stream (640 kbps, 8 kHz). I won't take into account the further 511 bytes
 * from the bit reservoir because they are expected to be in previous blocks (if any).
 * Further, due to `GetInfoTag' seeking for a new sync bit sequence, the buffer has to be
 * large enough to contain a full frame and a new mpeg header (4 bytes).
 */

#define LARGEST_BUFFER 5765

// In the worst case, 9 bitstream frames have to be received before decoding can start.
#define NP23B (9+3)

/*
 * This structure will store the bytes for part2 (scalefactors)
 * and part3 (Huffman encoded data) of the main data
 */
typedef struct PART23BUF {
    unsigned char buf[LARGEST_BUFFER];
    int len;
    off_t pos;
} PART23BUF;

typedef struct FILEBITS {
    FILE         *fp;
    unsigned int lastBits;
    int          nLastBits;
    off_t        location;
    off_t        latest_bit; /* offset of the latest bit */
} FILEBITS;

void initBits(FILEBITS *fb,FILE *fp, off_t eobs);
off_t ftellBits(FILEBITS *fb);
void fseekBits(FILEBITS *fb,off_t offset,int whence);
unsigned long int readBits(FILEBITS *fb,int n);
char seek_p23b(off_t pos);
int skip_p23b(int len);
int skipback_p23b(int len);
off_t tell_p23b(void);
void load_into_p23b(FILEBITS *fb,int len);
unsigned char readOneByte_p23b(void);

#endif
