/*
 *   mp3guessenc header
 *   Copyright (C) 2002-2010 Naoki Shibata
 *   Copyright (C) 2011-2013 Elio Blanca <eblanca76@users.sourceforge.net>
 *
 * Xing VBR tagging for LAME.
 * Copyright (c) 1999 A.L. Faber
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef MP3GUESSENC_H
#define MP3GUESSENC_H

#include "tags.h"

#define HEADER_FIELD_SYNC      0xFFE00000
#define HEADER_FIELD_MPEG_ID     0x180000
#define HEADER_FIELD_LSF          0x80000 /* this is not a *real* field, it is just a commodity for accessing lsf bit */
#define HEADER_FIELD_LAYER        0x60000
#define HEADER_FIELD_CRC          0x10000
#define HEADER_FIELD_BITRATE       0xF000
#define HEADER_FIELD_SAMPRATE       0xC00
#define HEADER_FIELD_PADDING        0x200
#define HEADER_FIELD_PRIVATE        0x100
#define HEADER_FIELD_CHANNELS        0xC0
#define HEADER_FIELD_MODEXT          0x30
#define HEADER_FIELD_COPYRIGHT        0x8
#define HEADER_FIELD_ORIGINAL         0x4
#define HEADER_FIELD_EMPHASIS         0x3

#define HEADER_CONSTANT_FIELDS_STRICT ( \
        HEADER_FIELD_SYNC |      \
        HEADER_FIELD_MPEG_ID |   \
        HEADER_FIELD_LAYER |     \
        HEADER_FIELD_CRC |       \
        HEADER_FIELD_CHANNELS |  \
        HEADER_FIELD_SAMPRATE |  \
        HEADER_FIELD_COPYRIGHT | \
        HEADER_FIELD_ORIGINAL)

#define HEADER_CONSTANT_FIELDS ( \
        HEADER_FIELD_SYNC |      \
        HEADER_FIELD_MPEG_ID |   \
        HEADER_FIELD_LAYER |     \
        HEADER_FIELD_CRC |       \
        HEADER_FIELD_SAMPRATE |  \
        HEADER_FIELD_COPYRIGHT | \
        HEADER_FIELD_ORIGINAL)

#define HEADER_ANY_BUT_BITRATE_AND_PADDING_FIELDS ( \
        HEADER_FIELD_SYNC |      \
        HEADER_FIELD_MPEG_ID |   \
        HEADER_FIELD_LAYER |     \
        HEADER_FIELD_CRC |       \
        HEADER_FIELD_SAMPRATE |  \
        HEADER_FIELD_PRIVATE |   \
        HEADER_FIELD_CHANNELS |  \
        HEADER_FIELD_MODEXT |    \
        HEADER_FIELD_COPYRIGHT | \
        HEADER_FIELD_ORIGINAL |  \
        HEADER_FIELD_EMPHASIS)

#define SCAN_SIZE 65536  /* default is to scan 64 kB when searching for a valid frame */
#define RESYNC_POS_CURRENT  ((off_t)-1) /* using resync_mp3(fp, &h, RESYNC_POS_CURRENT) makes it start its
                                * search from current position, instead of fseeking to `pos' */

typedef struct detectionInfo {
    int     blockCount[4];
    int     modeCount[5];
    char    usesScfsi;         /* this must be set to 0 */
    char    usesScalefacScale; /* this must be set to 0 */
    char    usesPadding;       /* this must be set to 0 */
    char    encoder_string[LAME_STRING_LENGTH];
    char    vbr_stream;
    char    vbr_tag;
    char    lameTag;
    char    frameCountDiffers;
    int     freq;
    int     ancillaryData;
    unsigned char   lay;       /* layer */
    unsigned char   id;        /* mpeg id (1/2/2.5) */
/*
 * mp3PRO detection
 * ----------------
 * The mp3PRO detection method searchs for a "signature" the encoder puts into its audio data.
 * The encoder stores higher frequency informations "hidden" into ancillary data of layerIII
 * streams while creating mpeg2 streams only. These infos are splitted into small packets (min
 * 6 bytes) and, given the first two bytes (big endian), they always show bits 15, 14 and 3 set.
 * Having said that, identification of the encoder is straightforward, since Fraunhofer IIS is
 * the patent owner and the only one providing encoders for mp3PRO.
 */
#define PRO_SIGN_BYTE1    0xC0
#define PRO_SIGN_BYTE2    0x08
/*
 * The detection for mp3Surround (5.1 stream into a backward compatible mpeg1 layerIII stream)
 * works the same way, just the signature is obviously different.
 */
#define SURR_SIGN_BYTE1   0xCF
#define SURR_SIGN_BYTE2   0x30
    unsigned char enhSignature[2]; /* these bytes of enhanced mp3 features must be set to 0xFF */
    char    ofl;
} detectionInfo;

typedef struct frameCounterCell {
    int frameSize;
    int frameCount;
} frameCounterCell;

typedef struct streamInfo {
    off_t           filesize;
    off_t           totFrameLen;  /* this must be set to 0 */
    int             totFrameNum;  /* this must be set to 0 */
    unsigned char   lsf;
    char            mode;
    char            crc;
    char            copyright;
    char            original;
    char            emphasis;
    int             reservoirMax; /* this must be set to 0 */
    char            freeformat;   /* this must be set to 0 */
    int             bitrateCount[15];
    frameCounterCell    unpadded[15];
    frameCounterCell    padded[15];
    int             nSyncError;
    int             ancill_min;
    int             ancill_max;
    short           encDelay;
    int             orig_samples;
    unsigned char   min_global_gain[2];
    unsigned char   max_global_gain[2];
} streamInfo;

extern int samples_tab[2][3];
extern int samplerate_tab [3][4];
extern int bitrate_tab[2][3][16];
extern char sideinfo_tab[2][2];

char head_check(int head, char relaxed);

#endif
