/*
 *   tags is a support module which provides access to metadata tags (xing/vbri/lame tag)
 *   Copyright (C) 2013 Elio Blanca <eblanca76@users.sourceforge.net>
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

/* this will enable 64 bit integers as file offset on 32 bit posix machines */
#define _FILE_OFFSET_BITS 64

#ifdef __MINGW32__
#define __USE_MINGW_FSEEK         /* Request mingw internal implementation of fseeko64 */
#define __USE_MINGW_ANSI_STDIO 1  /* Select a more ANSI C99 compatible implementation of printf() and friends. */
#endif /* __MINGW32__ */

#include <ctype.h>
#include <stdio.h>
#include <string.h>

#ifdef __MINGW32__
#define off_t     off64_t
#define fseeko    fseeko64
#define ftello    ftello64
#endif /* __MINGW32__ */

#include "tags.h"
#include "mp3guessenc.h"

int genre_last=147;
char *genre_list[]={
    "Blues", "Classic Rock", "Country", "Dance", "Disco", "Funk",
    "Grunge", "Hip-Hop", "Jazz", "Metal", "New Age", "Oldies",
    "Other", "Pop", "R&B", "Rap", "Reggae", "Rock",
    "Techno", "Industrial", "Alternative", "Ska", "Death Metal", "Pranks",
    "Soundtrack", "Euro-Techno", "Ambient", "Trip-Hop", "Vocal", "Jazz+Funk",
    "Fusion", "Trance", "Classical", "Instrumental", "Acid", "House",
    "Game", "Sound Clip", "Gospel", "Noise", "AlternRock", "Bass",
    "Soul", "Punk", "Space", "Meditative", "Instrumental Pop", "Instrumental Rock",
    "Ethnic", "Gothic", "Darkwave", "Techno-Industrial", "Electronic", "Pop-Folk",
    "Eurodance", "Dream", "Southern Rock", "Comedy", "Cult", "Gangsta",
    "Top 40", "Christian Rap", "Pop/Funk", "Jungle", "Native American", "Cabaret",
    "New Wave", "Psychadelic", "Rave", "Showtunes", "Trailer", "Lo-Fi",
    "Tribal", "Acid Punk", "Acid Jazz", "Polka", "Retro", "Musical",
    "Rock & Roll", "Hard Rock", "Folk", "Folk/Rock", "National Folk", "Swing",
    "Fast-Fusion", "Bebob", "Latin", "Revival", "Celtic", "Bluegrass", "Avantgarde",
    "Gothic Rock", "Progressive Rock", "Psychedelic Rock", "Symphonic Rock", "Slow Rock", "Big Band",
    "Chorus", "Easy Listening", "Acoustic", "Humour", "Speech", "Chanson",
    "Opera", "Chamber Music", "Sonata", "Symphony", "Booty Bass", "Primus",
    "Porn Groove", "Satire", "Slow Jam", "Club", "Tango", "Samba",
    "Folklore", "Ballad", "Power Ballad", "Rhythmic Soul", "Freestyle", "Duet",
    "Punk Rock", "Drum Solo", "A capella", "Euro-House", "Dance Hall",
    "Goa", "Drum & Bass", "Club House", "Hardcore", "Terror",
    "Indie", "BritPop", "NegerPunk", "Polsk Punk", "Beat",
    "Christian Gangsta", "Heavy Metal", "Black Metal", "Crossover", "Contemporary C",
    "Christian Rock", "Merengue", "Salsa", "Thrash Metal", "Anime", "JPop",
    "SynthPop",
};

// Utility routines

unsigned int ExtractI4(char *buf)
{
    unsigned int x;
    x = (unsigned char)buf[0];
    x <<= 8;
    x |= (unsigned char)buf[1];
    x <<= 8;
    x |= (unsigned char)buf[2];
    x <<= 8;
    x |= (unsigned char)buf[3];
    return x;
}

unsigned short ExtractI2(char *buf)
{
    unsigned short x;
    x = (unsigned char)buf[0];
    x <<= 8;
    x |= (unsigned char)buf[1];
    return x;
}

int extract_enc_string(char *lame_string,char *p,int len)
/*
 * here the search for lame extended version string is performed
 * lame stores this string into ancillary data since version 3.80b (may 6th 2000 - the previous
 * lame version 3.70 apr 6 2000 didn't have it)
 * Return values:
 * - lame_string contains lame extended version string (it is untouched when not found)
 * Note: this routine may be called several times since ancillary data may be found in
 * several frames, expecially in CBR files
 * it overwrites `lame_string' several times in order to fullfill it with data from the longest
 * of the strings it finds.
 */
{
//#define SHOW_LAME_STRINGS  /* useful when debugging */
    int i,len_before;
    char *q;

#ifdef SHOW_LAME_STRINGS
    printf("begin - lame_string=`%s', len=%d\n",lame_string,len);
    for (i=0; i<len; i++)
        if (isprint(p[i])) putchar(p[i]); else putchar('.');
    printf("\n");
#endif

    q = strstr(p,"LAME");
    if (q == NULL)
    {
        q = strstr(p,"GOGO");
        if (q != NULL)
        {
            strcpy(lame_string,"GOGO");
            return 4;
        }
        else
        {
#ifdef SHOW_LAME_STRINGS
            printf("No_lame_string.\n");
#endif
            return 0;
        }
    }
    len -= (q-p);

    if (lame_string[0] == '\0')
        strcpy(lame_string,"LAME"); /* this is likely to be the first call to extract_lame_string */
    else
    {
        if (lame_string[0]!='L' || lame_string[1]!='A' || lame_string[2]!='M' || lame_string[3]!='E')
        { /* lame_string is only modified here, it MUST be "" (null string) the first time,
             or it MUST start with "LAME", in any other case I want to be warned */
            printf("WARNING: buffer overwrite on lame_string! RESETTING!\n");
            memset(lame_string, 0, LAME_STRING_LENGTH);
            strcpy(lame_string,"LAME");
        }
    }

    i=4;   /* I'm sure q points to `LAME' so the cycle can be shortened */
    len_before=strlen(lame_string);

    while (i<len && i<LAME_STRING_LENGTH-1 && isprint(q[i]) && q[i]!='L' && q[i]!=0x55 && (unsigned char)q[i]!=0xAA)
    /* in recent versions, lame strings may end either with a 0x55 or a 0xAA byte */
    /* older releases used to put random non-printable bytes */
    /* Also, I check for a capital `L' in order to skip string repetitions such as `LAME3LAME3' */
    {
        lame_string[i]=q[i];
        i++;
    }

    if (i>=len_before) lame_string[i]=0;
#ifdef SHOW_LAME_STRINGS
    printf("exiting - lame_string=%s\n",lame_string);
#endif
    return i;
}

///////////////////////////////////////////////////////
// VBR TAG related routines

void reset_tag_data(VBRTAGDATA *p)
{
    p->alert=0;
    p->tagId=NULL;
    p->tagStartsAt=0;
//    p->tagEndsAt=0;
    p->header=0;
    p->frameSize=0;
    p->infoTag=TAG_NOTAG;
    p->version=0;
    p->encDelay=0;
    p->encPadding=0;
    p->tocEntries=0;
    p->sizePerTocEntry=0;
    p->framesPerTocEntry=0;
    p->tocSize=0;
    p->frames=0;
    p->bytes=0;
    p->vbr_scale=-1;  /* GOGO encoder does not put vbr quality value into its vbr tag, so I need here a flag */
    p->lame_buggy_vbrheader=0;
    memset(p->lametag, 0, LAMETAGSIZE);
}

int parse_vbri_tag(VBRTAGDATA *p, char *buf)
{
    int offset=4;

    p->tagId="VBRI";
    p->infoTag=TAG_VBRITAG;
    p->version=ExtractI2(buf+offset);
    offset+=2;
    p->encDelay=ExtractI2(buf+offset);
    offset+=2;
    p->vbr_scale=(int)ExtractI2(buf+offset);
    offset+=2;
    p->bytes=ExtractI4(buf+offset);
    offset+=4;
    p->frames=ExtractI4(buf+offset);
    offset+=4;
    p->tocEntries=ExtractI2(buf+offset);
    offset+=2;
    //p->scaleF4TocEntries=ExtractI2(buf+offset);
    offset+=2;
    p->sizePerTocEntry=ExtractI2(buf+offset);
    offset+=2;
    p->framesPerTocEntry=ExtractI2(buf+offset);
    offset+=2;
    p->tocSize=(int)p->sizePerTocEntry*(int)p->tocEntries;
    offset+=p->tocSize;

//    p->tagEndsAt+=4+2+2+2+4+4+2+2+2+p->tocSize;

    return offset;
}

int parse_xing_tag(VBRTAGDATA *p, char *buf, char *lame_string)
{
#define XING_FLAG_FRAMES     0x0001
#define XING_FLAG_BYTES      0x0002
#define XING_FLAG_TOC        0x0004
#define XING_FLAG_VBR_SCALE  0x0008

    int offset=4,flags;

    p->tagId = "Xing";
    if (buf[0]=='I') p->infoTag = TAG_LAMECBRTAG;
    else p->infoTag = TAG_XINGTAG;

    flags = ExtractI4(buf+offset);
    offset += 4;
//    p->tagEndsAt+=4;

    if (flags & XING_FLAG_FRAMES)
    {
        p->frames = ExtractI4(buf+offset);
        offset += 4;
//        p->tagEndsAt+=4;
    }

    if (flags & XING_FLAG_BYTES)
    {
        p->bytes = ExtractI4(buf+offset);
        offset += 4;
//        p->tagEndsAt+=4;
    }

    if (flags & XING_FLAG_TOC)
    {
        /*for(i=0;i<NUMTOCENTRIES;i++)
            store_the_toc();*/
        p->tocEntries = 100;
        p->sizePerTocEntry = 1;
        p->tocSize = 100;
        offset+=100;
//        p->tagEndsAt+=100;
    }

    if (
    buf[offset] == 'G' && buf[offset+1] == 'O' && buf[offset+2] == 'G' && buf[offset+3] == 'O'
    )
    {   /* gogo seems not to put a vbr quality value into its tag */
        offset += extract_enc_string(lame_string,buf+offset,4);
    }
    else
        if (flags & XING_FLAG_VBR_SCALE)
        {
            p->vbr_scale = ExtractI4(buf+offset);
            offset+=4;
//            p->tagEndsAt+=4;
        }

/*
 * Starting with lame-3.99 alpha releases the devs changed the lame tag signature so it didn't
 * start with the string `LAME', instead it just contained the capital `L' at the beginning and the
 * following expected characters in the form `Lx.xxyzz', where x is a digit and z is an optional digit.
 * Anyway the old signature `LAMEx.xxy' (y may be one of a/b/r) was restored with lame 3.99.2 release
 * in order to keep lame tag readable by old sw/hw players.
 */
    if (
    (buf[offset] == 'L' && buf[offset+1] == 'A' && buf[offset+2] == 'M' && buf[offset+3] == 'E')
    ||
    (buf[offset] == 'L' && isdigit(buf[offset+1]) && buf[offset+2] == '.' && isdigit(buf[offset+3]) && isdigit(buf[offset+4]))
    )
    {
        /*
         * maybe we found a lame info tag
         * warn: a version check has to be done because lame 3.90 was the first
         * version able to write a real info tag.
         * previous versions wrote just a string
         * Updated `if' statement for better handling of upcoming 3.100 release
         */
//        if (buf[offset+1]=='A' && buf[offset+4]=='3' && buf[offset+6]<'9')
        if (buf[offset+1]=='A' && buf[offset+4]=='3' && buf[offset+5]=='.' && isdigit(buf[offset+6]) && buf[offset+6]<'9' && isdigit(buf[offset+7]) && !isdigit(buf[offset+8]))
            /* old versions had more room for detailed strings due to lack of info tag */
            offset += extract_enc_string(lame_string,buf+offset,50);
        else
        {
            memcpy(p->lametag, buf+offset, LAMETAGSIZE);
            extract_enc_string(lame_string,buf+offset,9);
            offset += LAMETAGSIZE;
        }
    }

    return offset;
}

char GetInfoTag(VBRTAGDATA *pTagData, char *buf, int br, int pos, char *enc_string)
{
/*
 * Here we detect the info vbr/cbr tag (if any) -- the search is done inside the buffer
 * Return values:
 *  0: no tag
 *  1: tag found (details into pTagData)
 */
    int head,temp;
    int offset=1; /* !!! If no tag is found, then offset=1 will be used as starting point for resync! (avoid seeking to the frame itself) */
    char ret=0;
    unsigned char mono,lsf,crc;

    reset_tag_data(pTagData);

    head=ExtractI4(buf); /* this mpeg header has already been checked before calling GetInfoTag */
    pTagData->header=head; /* store the header */
    lsf  = 1-((head&HEADER_FIELD_LSF)>>19);
    mono =((head&HEADER_FIELD_CHANNELS)>>6)/3;
    crc  = 1-((head&HEADER_FIELD_CRC)>>16);

    if ((buf[36]=='V')&&(buf[37]=='B')&&(buf[38]=='R')&&(buf[39]=='I'))
    {
//        pTagData->tagEndsAt=40;     /* of course this value will be updated into `parse_vbri_tag' */
        offset = 36 + parse_vbri_tag(pTagData,buf+36);
    }
    else
    {
        temp = 4+(int)sideinfo_tab[mono][lsf]; /* placement of xing vbr tag doesn't take into account optional crc16 */

        if (
        ((buf[temp] == 'X') && (buf[temp+1] == 'i') && (buf[temp+2] == 'n') && (buf[temp+3] == 'g'))
        ||
        ((buf[temp] == 'I') && (buf[temp+1] == 'n') && (buf[temp+2] == 'f') && (buf[temp+3] == 'o'))
        )
        {
//            pTagData->tagEndsAt=offset+4;     /* of course this value will be updated into `parse_xing_tag' */
            offset = temp + parse_xing_tag(pTagData,buf+temp,enc_string);
            pTagData->lame_buggy_vbrheader=(((pTagData->header&HEADER_FIELD_BITRATE)>>12)==0x0f);
            if (pTagData->lame_buggy_vbrheader)
                pTagData->header = pTagData->header & ~HEADER_FIELD_BITRATE;
        }
    }

    if (pTagData->infoTag==TAG_NOTAG && ((pTagData->header&HEADER_FIELD_LAYER)>>17)==1 ) /* the second check is for layerIII */
    {   /* no tag was found but the actual frame size can be retrieved anyway */
        /* In fact, this is a layerIII frame and I know how to reach the end of meaningful data */
        /* this process will not require to follow a backpointer because I'm going to read the first frame */
        #define TWELVE_BIT_MASK 0x0FFF
        int part23_len=0,value;
        if (lsf)
        {
            if (mono)
            {   /* single value, offset 9 bits (1B+1b) */
                value=ExtractI4(buf+4+2*crc+0); /* this value falls entirely into the first 4 bytes block */
                value=(value>>11)&TWELVE_BIT_MASK;
                part23_len+=value;
            }
            else
            {   /* two values, offset 10 bits (1B+2b) and 73 bits (9B+1b) */
                value=ExtractI4(buf+4+2*crc+0); /* this value falls entirely into the first 4 bytes block */
                value=(value>>10)&TWELVE_BIT_MASK;
                part23_len+=value;
                value=ExtractI4(buf+4+2*crc+9);
                value=(value>>19)&TWELVE_BIT_MASK;
                part23_len+=value;
            }
        }
        else
        {
            if (mono)
            {   /* two values, offset 18 bits (2B+2b) and 77 bits (9B+5b) */
                value=ExtractI4(buf+4+2*crc+0); /* this value falls entirely into the first 4 bytes block */
                value=(value>>2)&TWELVE_BIT_MASK;
                part23_len+=value;
                value=ExtractI4(buf+4+2*crc+9);
                value=(value>>15)&TWELVE_BIT_MASK;
                part23_len+=value;
            }
            else
            {   /* four values, offset 20 bits (2B+4b), 79 bits (9B+7b),
                   138 bits (17B+2b) and 197 bits (24B+5b) */
                value=ExtractI4(buf+4+2*crc+0); /* this value falls entirely into the first 4 bytes block */
                value=value&TWELVE_BIT_MASK;
                part23_len+=value;
                value=ExtractI4(buf+4+2*crc+9);
                value=(value>>13)&TWELVE_BIT_MASK;
                part23_len+=value;
                value=ExtractI4(buf+4+2*crc+17);
                value=(value>>18)&TWELVE_BIT_MASK;
                part23_len+=value;
                value=ExtractI4(buf+4+2*crc+24);
                value=(value>>15)&TWELVE_BIT_MASK;
                part23_len+=value;
            }
        }
        /* now I've got to find the sync bits of the next frame - if any - */
        /* in order to know the actual size of this frame */
        /* I'll start my search from the end of last part3 on, */
        /* so sure I will skip all the meaningful data */
        offset=4+2*crc+(int)sideinfo_tab[mono][lsf]+part23_len/8;
    }

    if (pTagData->infoTag!=TAG_NOTAG) /* was any tag found? If so, I am sure I'm dealing with a layerIII stream */
    {
        pTagData->tagStartsAt=pos;  /* offset of the mpeg frame containing this tag */
        pTagData->alert=(offset>br);/* if the offset of the last byte parsed is greater than bytes actually read, then set alert for parsing */
        ret = 1;
    }

    if (pTagData->alert)
        pTagData->frameSize = br;
    else
    {
        /* now I've got to find the sync bits of the next frame - if any - */
        /* this way I can calculate the length of the actual frame */
        /* these statements can also be reached by non-layerIII streams */
        head  = (unsigned char)buf[offset];   head = head << 8;
        head |= (unsigned char)buf[offset+1]; head = head << 8;
        head |= (unsigned char)buf[offset+2]; head = head << 8;
        while (offset<br-3)
        {
            head |= (unsigned char)buf[offset+3];
            if (head_check(head,0))
            {
                /* when seeking for the next valid mpeg header, special care must be taken */
                /* after a freeformat frame, I expect a new freeformat frame */
                /* also, after a `regular' bitrate frame, a `regular' bitrate frame has to be expected */
                if ((pTagData->header&HEADER_FIELD_BITRATE)==0 && (head&HEADER_FIELD_BITRATE)==0)
                    break;
                if ((pTagData->header&HEADER_FIELD_BITRATE)!=0 && (head&HEADER_FIELD_BITRATE)!=0)
                    break;
            }
            head = head << 8;
            offset++;
        }
        pTagData->frameSize=offset;    /* save the frame size */
    }

    return ret;
}

void show_info_tag (VBRTAGDATA *p)
/*
 * Here we show the technical details found into the VBRI/XING tag
 * If a lame tag is found, then it's showed as well.
 */
{
    unsigned char lameTag=(p->lametag[0]!=0);

    //if (p.alert) printf("Found an mpeg tag but an error occurred. Details may be inaccurate.\n");
    printf("%s tag detected into the first frame (%d bytes long).\n",p->tagId,p->frameSize);
    printf("  Tag offset       : %lld (0x%08X)\n",(long long)p->tagStartsAt,(unsigned)p->tagStartsAt);
    if (p->infoTag==TAG_VBRITAG)
    {
        printf("  Tag version      : %d\n",p->version);
        printf("  Encoder delay    : %d samples\n",p->encDelay);
    }
    printf("  File size        : %lu bytes\n",p->bytes);
    printf("  Number of frames : %u\n",p->frames);
    if (p->vbr_scale!=-1) /* GOGO doesn't add a vbr quality value */
    {
        printf("  Quality          : %u",p->vbr_scale);
        if ((lameTag)&&(p->infoTag!=TAG_LAMECBRTAG))
        { /* sure the file is vbr encoded by lame, I just need to know whether it's abr or vbr */
            unsigned char lamemode=p->lametag[9]&15;
            if ((lamemode>2)&&(lamemode<7))
                /* it's vbr, so let's print encoder options */
                printf(" (-q %d -V %d)",(100-p->vbr_scale)%10,(100-p->vbr_scale)/10);
        }
        printf("\n");
    }
    printf("  TOC              : ");
    if (p->tocSize) printf("%d bytes (%d entries, %d byte%s each)\n",p->tocSize,p->tocEntries,p->sizePerTocEntry,(p->sizePerTocEntry==1)?"":"s");
    else printf("no\n");
    if (p->infoTag!=TAG_VBRITAG)
    {
        printf("  Lame tag         : ");
        if (lameTag)
        {
            unsigned char c,i;
            unsigned short j;
            unsigned int delay;
            printf("yes");
            if (p->lame_buggy_vbrheader) printf(" (buggy bitrate field)");
            /* print details from lame tag */
            printf("\nLame tag details...\n  Lame short string     : ");
            for(i=0;i<9;i++)
                if (isprint(p->lametag[i])) putchar(p->lametag[i]); else putchar(' ');
            c=p->lametag[9]>>4;
            printf("\n  Tag revision          : ");
            if (c!=15) printf("%d",c);
            else printf("invalid!");
            c=p->lametag[9]&15;
            printf("\n  Bitrate strategy      : ");
            if ((c==2)||(c==9))
                printf("ABR, ");  /* it's abr */
            else
            {
                if ((c==1)||(c==8))
                    printf("CBR, ");  /* it's cbr */
                else
                    if ((c>2)&&(c<7)) /* it's vbr */
                        printf("VBR method %s, min ",(c==3)?"old/rh":(c==4)?"mtrh":/*if lame, c is 5*/"mt");
            }
            c=p->lametag[20];
            if (c)
            {
                printf("%d kbps",c);
                if (c==255) printf (" or higher");
            }
            else
                printf("unknown");
            c=p->lametag[10];
            printf("\n  Lowpass value         : ");
            if (c) printf("%d",c*100);
            else printf("unknown");
            c=p->lametag[19];
            printf("\n  nspsytune             : %s\n",(c&0x10)?"yes":"no");
            printf("  nssafejoint           : %s\n",(c&0x20)?"yes":"no");
            printf("  nogap continued       : %s\n",(c&0x40)?"yes":"no");
            printf("  nogap continuation    : %s\n",(c&0x80)?"yes":"no");
            printf("  ATH type              : %d\n",c&0xf);
            delay=ExtractI4(p->lametag+20);
            p->encDelay=(short)((delay>>12)&0xfff);
            p->encPadding=(short)(delay&0xfff);
            printf("  Encoder delay (start) : %d samples\n",p->encDelay);
            printf("  Encoder padding (end) : %d samples\n",p->encPadding);
            c=p->lametag[24];
            printf("  Encoding mode         : ");
            switch (c&0x1c)
            {
                case 24:
                    printf("intensity stereo");
                    break;
                case 20:
                    printf("auto");
                    break;
                case 16:
                    printf("forced MS stereo");
                    break;
                case 12:
                    printf("joint stereo");
                    break;
                case 8:
                    printf("dual channel");
                    break;
                case 4:
                    printf("simple LR stereo");
                    break;
                case 0:
                    printf("mono");
                    break;
                default:
                    printf("other");
            }
            printf("\n  Unwise settings       : %sused\n",(c&32)?"":"not ");
            printf("  Source frequency      : ");
            switch (c&0xc0)
            {
                case 128:
                    printf("48 kHz");
                    break;
                case 64:
                    printf("44.1 kHz");
                    break;
                case 0:
                    printf("32 kHz or below");
                    break;
                default:
                    printf("higher than 48 kHz");
            }
            j=ExtractI2(p->lametag+26)&0x7ff;  /* 11 least significant bits are used for preset */
            printf("\n  Preset                : ");
            if (j==0)
                printf("No preset.");
            else
            {
                if (j<321)
                    printf("%d kbps",j);
                else
                {
                    if (j>409 && j<501)
                        printf("V%d / VBR_%d",50-j/10,j-400);
                    else
                    {
                        if (j>999 && j<1008)
                        {
                            switch (j)
                            {
                                case 1000:
                                    printf("R3mix.");
                                    break;
                                case 1001:
                                    printf("Standard.");
                                    break;
                                case 1002:
                                    printf("Extreme.");
                                    break;
                                case 1003:
                                    printf("Insane.");
                                    break;
                                case 1004:
                                    printf("Standard fast.");
                                    break;
                                case 1005:
                                    printf("Extreme fast.");
                                    break;
                                case 1006:
                                    printf("Medium.");
                                    break;
                                case 1007:
                                    printf("Medium fast.");
                                    break;
                            }
                        }
                        else
                            printf("Unknown preset.");
                    }
                }
            }
            printf("\n  Originally encoded    : %u bytes\n",ExtractI4(p->lametag+28));
        }
        else
            printf("no\n");
    }
    printf("\n");
}

///////////////////////////////////////////////////////

void show_id3v1(id3tag id3)
{
    unsigned char i,id3v11=(!id3.comment[28]&&id3.comment[29]);

    printf("  Title   : ");
    for(i=0;i<30;i++)
        if (isprint(id3.title[i])) putchar(id3.title[i]); else putchar('.');

    printf("\n  Artist  : ");
    for(i=0;i<30;i++)
        if (isprint(id3.artist[i])) putchar(id3.artist[i]); else putchar('.');

    printf("\n  Album   : ");
    for(i=0;i<30;i++)
        if (isprint(id3.album[i])) putchar(id3.album[i]); else putchar('.');

    printf("\n  Year    : ");
    for(i=0;i<4;i++)
        if (isprint(id3.year[i])) putchar(id3.year[i]); else putchar('.');

    printf("\n  Comment : ");
    if (id3v11)
    {
        for(i=0;i<28;i++)
            if (isprint(id3.comment[i])) putchar(id3.comment[i]); else putchar('.');
        printf("\n  Track # : %u", id3.comment[29]);
    }
    else
        for(i=0;i<30;i++)
            if (isprint(id3.comment[i])) putchar(id3.comment[i]); else putchar('.');

    printf("\n  Genre   : ");
    if (id3.genre > genre_last)
        printf("unknown");
    else
        printf("%s",genre_list[id3.genre]);
    printf("\n\n");

    return;
}

off_t checkid3v1(FILE *fi, id3tag *id3)
/*
 * Here the info tag id3v1.x is detected
 * This function returns either the offset of the tag
 * or the offset of EOF if no tag is found,
 * so the main cycle can stop scanning when this value has been reached
 * Note: the file pointer IS modified and then it IS restored
 */
{
    off_t filepos=ftello(fi),pos;
    int read;
    fseeko(fi,0,SEEK_END);
    pos = ftello(fi);

    fseeko(fi,-128,SEEK_END);
    read = fread(id3,sizeof(id3tag),1,fi);
    if (read==1)
    {
        if ((id3->tag[0] == 'T' && id3->tag[1] == 'A' && id3->tag[2] == 'G')
            ||
            (id3->tag[0] == 't' && id3->tag[1] == 'a' && id3->tag[2] == 'g'))
            pos -= 128;
    }
    fseeko(fi,filepos,SEEK_SET);
    return pos;
}

int checkid3v2(FILE *fi, unsigned char *idMaj, unsigned char *idmin)
{
/*
 * Here the info tag id3v2 is detected
 * This function returns the size of the tag (if detected)
 * so the main cycle can choose if either show it or skip it.
 * Also, idMaj and idmin provide id3V2 major and minor version number.
 * If the tag is not found, return value is 0.
 * If errors are detected, return value is -1.
 * Note: this routine DOES read data from the input file
 * but it DOES restore the file position before exiting.
 */

    off_t filepos=ftello(fi);
    int id3v2size=0;
    unsigned char c1,c2,c3,c4,c5;
    fseeko(fi,0,SEEK_SET);
    c1 = getc(fi);
    c2 = getc(fi);
    c3 = getc(fi);
    c4 = getc(fi);
    c5 = getc(fi);

    if (c1 == 'I' && c2 == 'D' && c3 == '3')
    {   /* has ID3v2 */
        *idMaj=c4; *idmin=c5;
        c5 = getc(fi);   /* get tag flags */
        c1 = getc(fi);
        c2 = getc(fi);
        c3 = getc(fi);
        c4 = getc(fi);
        if (!(c1&0x80) && !(c2&0x80) && !(c3&0x80) && !(c4&0x80) && !(c5&0x0f))
        {
            id3v2size = c1*2097152+c2*16384+c3*128+c4 +10; /* 10 is due to tag header (10 bytes long) */
            //if (c5&0x10) id3v2size+=10;  /* is tag footer present? */
        }
        else
            id3v2size=-1;
    }

    fseeko(fi,filepos,SEEK_SET);
    return id3v2size;
}

/*
 * check for ape tag at the end of the file
 * when placed at the end of the file (between the very last frame and the id3v1 tag)
 * ape tag can be both 1 and 2. Tag version is returned into `ape_vers', if found.
 * search is started at file position `pos' when end of mpeg stream is expected to be.
 * return value is tag length if found, else zero.
 * file pointer gets modified
 */
int checkapetagx(FILE *fi, off_t pos, int *ape_vers, int *ape_items)
{
    unsigned char footer[32];
    int ape_len=0;

    if ((pos-32)>=0)
    {
        fseeko(fi, pos-32, SEEK_SET);
        if (fread(footer,1,32,fi)==32)
        {
            if (footer[0]=='A' && footer[1]=='P' && footer[2]=='E' && footer[3]=='T' && footer[4]=='A' && footer[5]=='G' && footer[6]=='E' && footer[7]=='X')
            {
                /* ape tag found! */
                *ape_vers=footer[8]+footer[9]*256+footer[10]*65536+footer[11]*16777216;
                ape_len=footer[12]+footer[13]*256+footer[14]*65536+footer[15]*16777216;
                *ape_items=footer[16]+footer[17]*256+footer[18]*65536+footer[19]*16777216;
                if (footer[23]&0x80)
                    /* tag contains a header */
                    ape_len+=32;
            }
        }
    }

    return ape_len;
}

