/*
 *   tags is a support module which provides access to metadata tags (xing/vbri/lame tag)
 *   Copyright (C) 2013 Elio Blanca <eblanca76@users.sourceforge.net>
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef TAGS_H
#define TAGS_H

#define LAMETAGSIZE        36
#define LAME_STRING_LENGTH 50

typedef struct id3tag {
    unsigned char tag[3];
    unsigned char title[30];
    unsigned char artist[30];
    unsigned char album[30];
    unsigned char year[4];
    unsigned char comment[30];
    unsigned char genre;
} id3tag;

/* structure to receive extracted header */
/* toc may be NULL*/
typedef struct
{
    char    alert;
    char    *tagId;
    off_t   tagStartsAt;
    int     header;         /* mpeg header of the frame containing the tag */
//    int     tagEndsAt;      /* I need to know where the tag actually ends -- DO I ?*/
    int     frameSize;      /* I want to keep the size of the frame containing the tag */
#define TAG_NOTAG      0
#define TAG_VBRITAG    1
#define TAG_XINGTAG    2
#define TAG_LAMECBRTAG 3
    char    infoTag;
    short   version;
//    float   delay;
    short   encDelay;       /* encoder delay (start) */
    short   encPadding;     /* encoder padding (samples added at the end of the wave) */
    short   tocEntries;
//    short   scaleF4TocEntries;
    short   sizePerTocEntry;
    short   framesPerTocEntry;
    int     tocSize;
    int     frames;         /* total bit stream frames from Vbr header data */
    unsigned long bytes;          /* total bit stream bytes from Vbr header data*/
    int     vbr_scale;      /* encoded vbr scale from Vbr header data*/
    char    lame_buggy_vbrheader;
    char    lametag[LAMETAGSIZE];
} VBRTAGDATA;

unsigned int ExtractI4(char *buf);
unsigned short ExtractI2(char *buf);
int extract_enc_string(char *enc_string, char *p, int len);
char GetInfoTag(VBRTAGDATA *pTagData, char *buf, int br, int pos, char *enc_string);
void show_info_tag (VBRTAGDATA *p);
void show_id3v1(id3tag id3);
off_t checkid3v1(FILE *fi, id3tag *id3);
int checkid3v2(FILE *fi, unsigned char *idMaj, unsigned char *idmin);
int checkapetagx(FILE *fi, off_t pos, int *ape_vers, int *ape_items);

#endif
